classdef EU_network < distribution_network
    
    properties
    end
    
    methods
        function obj = EU_network(net_para)
            obj.net_para=net_para;
            
            branches_para = get_branch();
            obj.energy_flow = get_energyflow();
            obj.vol_flow = get_volflow();
            obj.vp_0 = get_vp_0();
            obj.v_ref = get_v_ref();
            inverter_para = get_inverter_para();
            obj.branches = obj.add_branch(branches_para,inverter_para);
            
            % disconnection manager
            obj.dis_man=discon_man(obj.v_ref(0),obj.branches);
            
            nxall = 0;
            nyall = 0;
            for i=1:numel(obj.branches)
                nxall = nxall+obj.branches{i}.get_nx;
                nyall = nyall+obj.branches{i}.get_ny;
            end
            obj.nxall = nxall;
            obj.nyall = nyall;
            
            % observers manager
            ob_tab = get_ob_tab();
            obj.obs_man = obs_man(ob_tab,obj);
            obj.dis_man.obs_man=obj.obs_man;
            initial_discon_obss(obj.dis_man,obj.net_para.d_para);
            
            obj.nall = obj.nxall+obj.nyall+obj.obs_man.n_obstates_all;
            
            idx_all.vs=[];idx_all.P=[];idx_all.Q=[];idx_all.q=[];
            idx_all.xhat=[];idx_all.zhat=[];
            for i=1:numel(obj.branches)
                bri=obj.branches{i};
                idx_all.vs=[idx_all.vs;bri.idx.vs];
                idx_all.P=[idx_all.P;bri.idx.P];
                idx_all.Q=[idx_all.Q;bri.idx.Q];
                idx_all.q=[idx_all.q;bri.idx.q];
            end
            for i=1:numel(obj.obs_man.obss)
                obsi=obj.obs_man.obss{i};
                idx_all.xhat=[idx_all.xhat;obsi.idx.xhat];
                idx_all.zhat=[idx_all.zhat;obsi.idx.zhat];
            end
            obj.idx_all=idx_all;
            
            % attack
            obj.get_attack(net_para.a_para);
            
            % reference signal
            obj.u_const=obj.set_input();
        end
        
        function get_attack(obj,attack_para)
            ut=make_input(obj);
            nu=numel(ut(0));
            attack_vec=zeros(nu,1);
            br_a=obj.branches{attack_para.bnum};
             
            obj.attack.ex=attack_para.attack_ex;
            
            obj.attack.t=attack_para.attack_t;
            if attack_para.attack_ex
                attack_vec(br_a.idx.v_ref)=attack_para.attack_amp;
            end
            
            obj.attack.attack_vec=attack_vec;
        end
    end
    
end

function branch_para = get_branch()
branch_array = [
    1 0.00343 0.04711 0 0 0 0 0;
    2 0.00172 0.02356 0 0 0 0 0;
    3 0.00343 0.04711 0 0 0 0 0;
    4 0.00515 0.07067 0 0 0 0 0;
    5 0.00172 0.02356 0 0 0 0 0;
    
    6 0.00147 0.02157 3500 2295 300 4200 1;
    7 0.00662 0.09707 5500 5440 960 6500 1;
    %7 0.00662 0.09707 1e1 1e1 1e1 1e1 1;
    8 0.00147 0.02157 4000 5440 480 4700 1;
    9 0.00147 0.02157 4500 2295 600 5300 1;
    %9 0.00147 0.02157 1e1 1e1 1e1 1e1 1;
    10 0.00147 0.02157 3000 2720 400 3600 1;
    ];
branch_para = array2table(branch_array, 'VariableNames', ...
    {'branch_num', 'R', 'X', 'rhog', 'rhoc', 'qc', 's', 'type'}); %type 0:noDG 1:DG
%branch_para{:,{'rhog','rhoc','qc'}} = 0*branch_para{:,{'rhog','rhoc','qc'}};
end

function branch_energyflow = get_energyflow()
flow_cell = cell(0);
flow_cell = [flow_cell; 2 -1 -6];
flow_cell = [flow_cell; 3 -2 -7];
flow_cell = [flow_cell; 4 -3 -8];
flow_cell = [flow_cell; 5 -4 -9];
flow_cell = [flow_cell; 0 -5 -10];

branch_energyflow = flow_cell;
end

function branch_voldropflow = get_volflow()
flow_array = [
    0 -1 -1
    1 -2 -2
    2 -3 -3
    3 -4 -4
    4 -5 -5
    -1 6 -6
    -2 7 -7
    -3 8 -8
    -4 9 -9
    -5 10 -10
    ];
branch_voldropflow = flow_array;
% v(1)^2+v(2)^2+beta(P(3),Q(3))
end

function inverter_para = get_inverter_para()
inverter_array = [
    6 1;
    7 1;
    8 1;
    9 1;
    10 1;
    ];
inverter_para = array2table(inverter_array, 'VariableNames', ...
    {'branch_num', 'tau'});
end

function vp_0 = get_vp_0()
%vp_0 = @(t) 0;
vp_0 = @(t) 230;
%vp_0 = @(t) 230+5*sin(10*t);
end

function v_ref = get_v_ref()
%v_ref = @(t) 0; % (Volt)
v_ref = @(t) 230; % (Volt)
%v_ref = @(t) 230+10*sin(10*t);
end

function ob_tab = get_ob_tab()
ob_array ={
%{
1 {[1 6]} {[1 6]};
2 {[2 7]} {[2 7]};
3 {[3 8]} {[3 8]};
4 {[4 9]} {[4 9]};
5 {[5 10]} {[5 10]};
%}

%{
1 {[1 6]} {[6]};
2 {[2 7]} {[7]};
3 {[3 8]} {[8]};
4 {[4 9]} {[9]};
5 {[5 10]} {[10]};
%}


1 {[1 6]} {[6]};
2 {[2 7]} {[7]};
3 {[3 4 5 8 9 10]} {[8 9 10]};


%{
1 {[1 2 6 7]} {[1 2 6 7]};
2 {[3 8]} {[3 8]}
3 {[4 5 9 10]} {[4 5 9 10]};
%}

%1 {[6]} {[6]};
%2 {[7]} {[7]};
%3 {[8]} {[8]};
%4 {[4 9]} {[4 9]};
%5 {[10]} {[10]};
    };
ob_tab = cell2table(ob_array, 'VariableNames', ...
    {'observer_num', 'branches_area', 'measured_branches'});
% measured output is assumed to be voltage
end
