classdef branch_DG_fo < branch
    
    properties
        tau
        K
        rhog_const
        rhoc_const
        qc_const
        rhog
        rhoc
        qc
        v_ref
        nx=1
        ny=1
        nu=1
        idx
        x
        s
        Qmax
        para_droop
        droop_ctrl
    end
    
    methods
        function obj = branch_DG_fo(b_para,inv_para,v_ref)
            obj@branch(b_para);
            rhog0=b_para{:,'rhog'};
            rhoc0=b_para{:,'rhoc'};
            qc0=b_para{:,'qc'};
            s0=b_para{:,'s'};
            tau0=inv_para;
            obj.tau=tau0;
            obj.rhog_const=rhog0;
            obj.rhog=obj.get_rhog();
            obj.rhoc_const=rhoc0;
            obj.rhoc=obj.get_rhoc();
            obj.qc_const=qc0;
            obj.qc=obj.get_qc();
            obj.v_ref=v_ref;
            
            obj.s=s0;
            obj.Qmax=sqrt(s0^2-obj.rhog_const-2);
            
            obj.para_droop.wn=0;
            obj.para_droop.wm=0;
            obj.para_droop.wmax=17320.5;
            obj.para_droop.wmin=-17320.5;
            obj.droop_ctrl.linear=true;
            obj.droop_ctrl.DZ_sat=false;
            obj.K=get_droop_gain(obj);
        end
        
        function f = get_fi(obj,var,u_refs)
            % u_ref: v_ref
            vs=var(1);
            qg=var(2);
            
            f = -qg/obj.tau+ctrl_input(obj,u_refs-vs)/obj.tau;
        end
        
        function u = ctrl_input(obj,y)
            if obj.droop_ctrl.linear
                u = obj.K.pos*y;
            end
            
            if obj.droop_ctrl.DZ_sat
                p_d = obj.para_droop;
                if y<p_d.wmin
                    u=-obj.Qmax;
                elseif y<p_d.wm
                    u=obj.K.neg*(y-p_d.wm);
                elseif y<p_d.wn
                    u=0;
                elseif y<p_d.wmax
                    u=obj.K.pos*(y-p_d.wn);
                else
                    u=obj.Qmax;
                end
            end
        end
        
        function K = get_droop_gain(obj)
            p_d = obj.para_droop;
            K.pos=obj.Qmax/(p_d.wmax-p_d.wn);
            K.neg=obj.Qmax/(p_d.wm-p_d.wmin);
        end
        
        function rhog = get_rhog(obj)
            rhog = @(t) obj.rhog_const;
        end
        
        function rhoc = get_rhoc(obj)
            rhoc = @(t) obj.rhoc_const;
        end
        
        function qc = get_qc(obj)
            qc = @(t) obj.qc_const;
        end
        
        function add_idx(obj,idx)
            obj.idx.vs=idx+1;
            obj.idx.q=idx+2;
            obj.idx.P=[];
            obj.idx.Q=[];
        end
        
        function add_idx_u(obj,idx_u)
            obj.idx.v_ref=idx_u+1;
            obj.idx.pg=idx_u+2;
            obj.idx.pc=idx_u+3;
            obj.idx.qc=idx_u+4;
        end
    end
    
end