classdef distribution_network < handle
    
    properties
        net_para
        
        nxall
        nyall
        num_noDG
        num_DGfo
        branches
        vp_0
        v_ref
        energy_flow
        vol_flow
        obs_man
        idx_q
        nall
        idx_all % vs,P,Q,q,xhat,zhat
        sigs
        attack
        dis_man
        u_const
        t_int=0.01
    end
    
    methods
        
        
        function br = add_branch(obj,b_tab,inv_tab)
            lb = height(b_tab);
            br = cell(lb,1);
            obj.num_noDG = 0;
            obj.num_DGfo = 0;
            invidx = inv_tab{:,'branch_num'};
            n_v0=1;
            idx=0; idx_u=n_v0; count_q=0;
            for i=1:lb
                switch b_tab{i,'type'}
                    case 0
                        obj.num_noDG=obj.num_noDG+1;
                        br{i}=branch_noDG(b_tab(i,:));
                        br{i}.add_idx(idx);
                        br{i}.add_idx_u(idx_u);
                        idx=idx+3; count_q=count_q+3;
                    case 1                     
                        obj.num_DGfo=obj.num_DGfo+1;
                        j=find(invidx==b_tab{i,'branch_num'});
                        br{i}=branch_DG_fo(b_tab(i,:),inv_tab{j,'tau'},obj.v_ref);
                        br{i}.add_idx(idx);
                        br{i}.add_idx_u(idx_u);
                        idx=idx+2; idx_u=idx_u+4; count_q=count_q+2; obj.idx_q=[obj.idx_q;count_q];
                end
            end            
        end
        
        function f = get_f(obj,t,var,u_refs,attack,discon)
            % u_ref: [v'_0^2; v_ref^2; rhog; rhoc; qg]
            if nargin < 4
                u_refs = obj.u_const;
            end
            
            % adding attack
            if nargin < 5
                [attack_vec, attack_t]=set_attack_zero(numel(u_refs));
            else
                attack_vec=attack.attack_vec;
                attack_t=attack.t;
            end
            u_refs_at = add_attack(t,u_refs,attack_vec,attack_t);
            
            % disconnection parameter
            if nargin < 6
                discon=false([numel(obj.branches),1]);
            end
            
            
            % disconnection
            var_dis = obj.dis_man.discon_var(var,discon);
            u_refs_dis = obj.dis_man.discon_u(u_refs,discon);
            u_refs_at_dis = obj.dis_man.discon_u(u_refs_at,discon);
            
            
           
            f = [];
            % for branches
            for i=1:numel(obj.branches)
                bri = obj.branches{i};
                vari = [var_dis(bri.idx.vs); var_dis(bri.idx.q)];
                f = [f; bri.get_fi(vari,u_refs_at_dis(bri.idx.v_ref))];
            end
            
            
            % for observers
            % attack is not injected into observers
            if ~isempty(obj.obs_man)
                obsm = obj.obs_man;
                
                %w = compute_what_funhandle(obj,t,var,u_refs);
                w = compute_what_linear(obj,t,var_dis,u_refs_dis,discon);
                %w = compute_what_linear(obj,t,var_dis,u_ref2_at,discon);
                %w = var; %%%% CAUTION: for debug
                
                
                % disconnection
                %w = obj.dis_man.discon_obs(w,discon);
                
                
                % compute f(x)
                for i=1:numel(obsm.obss)
                    obsi = obsm.obss{i};
                    vari = var_dis([obsi.idx.xhat;obsi.idx.zhat]);
                   
                    int_sig = w(obsi.idx.ot);
                    %int_sig = var(obsi.idx.ot); %%%% CAUTION: for debug
                    meas_sig = var_dis(obsi.idx.ym); % measurement output
                    
                    f = [f; obsi.get_fi(vari,int_sig,meas_sig,u_refs)];
                end
                
            end
        end
        
        function g = get_g(obj,t,var,u_refs,attack,discon)
            % u_ref = [v'_0^2, v_ref^2, rhog; rhoc; qc];
            if nargin < 4
                u_refs= obj.u_const;
            end
            
            % adding attack
            if nargin < 5
                [attack_vec,attack_t]=set_attack_zero(numel(u_refs));
            else
                attack_vec=attack.attack_vec;
                attack_t=attack.t;
            end
            u_refs=add_attack(t,u_refs,attack_vec,attack_t);
            
            
            % disconnection parameter
            if nargin < 6
                discon=false([numel(obj.branches),1]);
            end
            
            % disconnection
            var_dis = obj.dis_man.discon_var(var,discon);
            u_refs_dis = obj.dis_man.discon_u(u_refs,discon);       
            
            br = obj.branches;
            ef = obj.energy_flow;
            vf = obj.vol_flow;
            nef = numel(ef); % # of nodal points
            nvf = size(vf,1); % # of branches
            
            g=zeros(2*nef+nvf,1);
            num_ui=4;
            num_v0=1;
            rhog = u_refs_dis(num_v0+2:num_ui:num_v0+2+(obj.num_DGfo-1)*num_ui);
            rhoc = u_refs_dis(num_v0+3:num_ui:num_v0+3+(obj.num_DGfo-1)*num_ui);
            qc = u_refs_dis(num_v0+4:num_ui:num_v0+4+(obj.num_DGfo-1)*num_ui);
            
            % conservation of energy
            for i=1:nef
                tmpf=obj.energy_flow{i};
                % ex [2 -1 -6] [ind of nodal point 1, *2, supplied power]
                for j=1:numel(tmpf)
                    ind = abs(tmpf(j));
                    if ind<=nef % P,Q variables
                        if ind~=0
                            bri = obj.branches{ind};
                            g(i) = g(i)+sign(tmpf(j))*var_dis(bri.idx.P);
                            g(nef+i) = g(nef+i)+sign(tmpf(j))*var_dis(bri.idx.Q);
                        end
                    else % given p,q
                        bri = obj.branches{ind};
                        idx_fo = ind-obj.num_noDG;
                        rhogtmp = rhog(idx_fo);
                        rhoctmp = rhoc(idx_fo);
                        rhotmp = rhogtmp-rhoctmp;
                        g(i) = g(i)+sign(tmpf(j))*rhotmp;
                        g(nef+i) = g(nef+i)+sign(tmpf(j))*(var_dis(bri.idx.q)-qc(idx_fo));
                    end
                end
            end
            
            % voltage drop
            for i=1:nvf
                tmpf=obj.vol_flow(i,:);
                ind=abs(tmpf);
                eps=1e-4;
                bris=cell(2,1);
                vtmp=zeros(2,1);
                for j=1:2
                    if ind(j)~=0
                        bris{j}=obj.branches{ind(j)};
                        vtmp(j)=var_dis(bris{j}.idx.vs);
                    else
                        vtmp(j)=u_refs_dis(1);
                    end
                end
                vartmp = sign(tmpf(1:2)+eps)'.*vtmp;
                bind = abs(tmpf(3));
                tmpb=br{bind};
                R=tmpb.R; X=tmpb.X;
                if bind<=nef
                    bri = obj.branches{bind};
                    var_vol = [vartmp; sign(tmpf(3))*beta_PQ(R,X,var_dis(bri.idx.P),var_dis(bri.idx.Q))];
                else
                    idx_fo = bind-obj.num_noDG;
                    rhotmp=rhog(idx_fo)-rhoc(idx_fo);
                    bri=obj.branches{bind};
                    qtmp=var_dis(bri.idx.q)-qc(idx_fo);
                    var_vol = [vartmp; sign(tmpf(3))*beta_PQ(R,X,rhotmp,qtmp)];
                end
                g(2*nef+i) = voltage_drop_eq(var_vol);
            end
            
            % disconnection
            g = obj.dis_man.discon_pow(g,var,nef,discon);
            g = obj.dis_man.discon_vol(g,var,u_refs_dis,nef,discon);
        end
        
        function out = eq_compute(obj,t,x0,u0)
            if nargin<2
                t=0;
            end
            if nargin<3
                ut = make_input(obj);
                u0 = ut(t);
                x0 = zeros(obj.nall,1);
            end
            options = optimoptions('fsolve');
            fg = @(x) [obj.get_g(t,x,u0); obj.get_f(t,x,u0)];
            out = fsolve(fg,x0,options);
        end
        
        function simulate(obj,t0,x0)
            discon=false([numel(obj.branches),1]);
            t_int0=obj.t_int;
            ut = make_input(obj);
            Mat = obj.make_Mat();
            fg = @(t,x) Mat*[obj.get_g(t,x,ut(t),obj.attack,discon); obj.get_f(t,x,ut(t),obj.attack,discon)];
            %fg = @(t,x) Mat*[obj.get_g(t,x,ut(t)); obj.get_f(t,x,ut(t))];
            nall0 = obj.nyall+obj.nxall;
            M = zeros(nall0);
            for i=1:nall0
                if any(ismember(obj.idx_q,i))
                    M(i,i)=1;
                end
            end
            M = blkdiag(M,eye(obj.obs_man.n_obstates_all));
            
            t_end=t0(1);
            t=[];y=[];
            t0_tmp=t0; x0_tmp=x0;
            eps=1e-6;
            while t_end<t0(end)
                if strcmp(obj.net_para.d_para.d_type,'detection')
                    op=odeset('Mass',M,'Events',@(t,y) obj.DisEventFcn_detect(t,y,ut(t),obj.attack,discon),'Maxstep',t_int0);
                elseif strcmp(obj.net_para.d_para.d_type,'time')
                    op=odeset('Mass',M,'Events',@(t,y) obj.DisEventFcn_time(t,y),'Maxstep',t_int0);
                end
                [t_tmp,y_tmp,t_end,y00all,dis_points] = ode15s(fg,t0_tmp,x0_tmp,op);
                t=[t;t_tmp]; y=[y;y_tmp];
                if t_end<t0(end)
                    % disconnection parameter
                    discon(dis_points)=true;
                    t_end=t_end(1);
                    obj.dis_man.discon_res.t=[obj.dis_man.discon_res.t, t_end+t_int0];
                    obj.dis_man.discon_res.discon=[obj.dis_man.discon_res.discon,discon];
                    
                    % calculating vs,P,Q after disconnection
                    q_tmp=y_tmp(end,obj.idx_q);
                    xhat_tmp=y_tmp(end,obj.obs_man.idx.xhat);
                    zhat_tmp=y_tmp(end,obj.obs_man.idx.zhat);
                    yFcn=@(x) obj.sub_xqxh(x,q_tmp,xhat_tmp,zhat_tmp);
                    gfun=@(x) obj.get_g(0,yFcn(x),ut(0),obj.attack,discon);
                    
                    idx_y0=1:obj.nall;
                    idxs=sort([obj.idx_q;obj.obs_man.idx.xhat;obj.obs_man.idx.zhat]);
                    idx_y0(idxs)=[];
                    y00=y00all(idx_y0);
                    %y00=zeros(obj.nyall,1);
                    y0=fsolve(gfun,y00);
                    
                    % setting new intial conditions
                    t0_tmp=[t_end+t_int0:t_int0:t0(end)]';
                    x0_tmp=sub_xqxh(obj,y0,q_tmp,xhat_tmp,zhat_tmp);
                    
                    % updating definition of fg with discon
                    fg = @(t,x) Mat*[obj.get_g(t,x,ut(t),obj.attack,discon); obj.get_f(t,x,ut(t),obj.attack,discon)];
                end
            end
            
            d_man=obj.dis_man;
            for i=1:numel(obj.branches)
                t_count=0;
                for j=1:size(t,1)
                    if numel(d_man.discon_res.t)==t_count
                        t_count=t_count-1;
                    end
                    if t(j)>d_man.discon_res.t(t_count+1)-1e-5
                        t_count=t_count+1;
                    end
                    if d_man.discon_res.discon(i,t_count)
                        bri = obj.branches{i};
                        y(j,bri.idx.vs)=NaN;
                        y(j,bri.idx.P)=NaN;
                        y(j,bri.idx.Q)=NaN;
                        y(j,bri.idx.q)=NaN;
                    end
                end
            end
            obj.sigs = obj.make_sigs(t,y);
        end
        
        function w = compute_what_linear(obj,t,var,u_refs,discon)
            % w = C(x,xhat)+B(u) (w: without q)
            % wq = Cqx
            % var: all variables
            
            w = zeros(obj.nall,1);
            obsm = obj.obs_man;
            C0 = obsm.obs_C;
            B0 = obsm.obs_B;
                       
            idx_x = obsm.idx.x;
            idx_xhat = obsm.idx.xhat;
            idx_zhat = obsm.idx.zhat;
            idx_y = obsm.idx.y;
            idx_ot = obsm.idx.ot;
            
            C=zeros(obj.nall,size(C0,2));
            B=zeros(obj.nall,size(B0,2));
            C(idx_y,:)=C0;
            B(idx_y,:)=B0;

            idx_y_dis = [];
            for i=1:numel(obj.branches)
                if discon(i)
                    bri=obj.branches{i};
                    idx_dis=[bri.idx.vs;bri.idx.P;bri.idx.Q];
                    for j=1:numel(idx_dis)
                        idx_y(idx_y==idx_dis(j))=[];
                        idx_y_dis=[idx_y_dis;idx_dis(j)];
                    end
                end
            end
            idx_y_dis=sort(idx_y_dis);
            
            w(idx_ot) = var(idx_ot);
            w(idx_x) = var(idx_x);
            w(idx_xhat) = var(idx_xhat);
            w(idx_zhat) = var(idx_zhat);
            w(idx_y_dis) = var(idx_y_dis);
            
            % \hat{w} = C1*int_sig+C2*\hat{w}+C3*\hat{x,z}+B*u
            %         = Cp*int_sig+Cp2*\hat{x,z}+Bp*u
            
            C1=C(idx_y,idx_ot); C2=C(idx_y,idx_y); C3=C(idx_y,[idx_x;idx_xhat;idx_zhat]);
            B1=B(idx_y,:);
            IiC2=inv(eye(numel(idx_y))-C2);
            Cp=IiC2*C1; Cp2=IiC2*C3; Bp=IiC2*B1;
            w_ot=Cp*w(idx_ot); %w_xhat=Cp2*w([idx_x;idx_xhat;idx_zhat]);
            w_xhat=Cp2*[w(idx_x);w(idx_xhat);w(idx_zhat)];
            
            C4=C(idx_y,idx_y_dis); w_ydis=(IiC2*C4)*w(idx_y_dis);
            
            % exception handling
            w_ot=emp_to_zero(w_ot);
            w_xhat=emp_to_zero(w_xhat);
            w_ydis=emp_to_zero(w_ydis);
            
            w(idx_y) = w_ot+w_xhat+w_ydis+Bp*u_refs;
        end
        
        function w = compute_what_funhandle(obj,t,var,u_refs)
            %%%
            %%% CAUTION: DON'T USE for LARGE-SCALE SYSTEMS
            %%%          SUPER HEAVY
            %%%
            obsm = obj.obs_man;
            
            idx_w = obsm.idx_w;
            idx_what = obsm.idx_what;

            count_hat=0; x=@(what) [];
            for j=1:obj.nxall+obj.nyall
                if any(ismember([idx_what],j))
                    count_hat = count_hat+1;
                    x = @(what) [x(what); what(count_hat)];
                else
                    x = @(what) [x(what); var(j)];
                end
            end

            Gam = @(what,u) [];
            what_perm = @(what) [];
            Ix = eye(obj.nxall+obj.nyall);
            for i=1:numel(obsm.obss)
                obsi = obsm.obss{i};
                Gam = @(what,u) [Gam(x(what),u); obsi.w_dyn(x(what),u)];

                [idx_x1, ~, idx_y1, ~] = idx_xy(obsi,obj);
                Ixwhat = Ix(idx_y1,:);
                Ixxwhat = Ix(idx_x1,:);
                what_perm = @(what) [what_perm(what); Ixwhat*x(what)];
                what_perm = @(what) [what_perm(what); Ixxwhat*x(what)];
            end

            Gam_all = @(what,u) what_perm(what)-Gam(what,u);
            Gamwhat = @(what) Gam_all(what,u_refs);
            est_what = fsolve(Gamwhat,zeros(numel(idx_what),1));
            w = zeros(obj.nxall+obj.nyall,1);
            w(idx_what) = est_what;
            w(idx_w) = var(idx_w);
        end
        
        function Mat = make_Mat(obj)
            cand1=[3*obj.num_noDG+1:3*obj.num_noDG+obj.num_DGfo];
            cand2=[3*obj.num_noDG+obj.num_DGfo+1:3*obj.num_noDG+2*obj.num_DGfo];
            
            change_table=1:3*obj.num_noDG;
            for i=1:obj.num_DGfo
                change_table=[change_table cand1(i) cand2(i)];
            end
            for i=1:obj.obs_man.n_obstates_all
                change_table = [change_table numel(change_table)+1];
            end
            
            I=eye(obj.nall);Mat=I;
            for i=1:obj.nall
                Mat(i,:) = I(change_table(i),:);
            end
        end
        
        function ut = make_input(obj)
            ut = @(t) obj.vp_0(t)^2;
            for i=1:obj.num_DGfo
                j=i+obj.num_noDG;
                ut = @(t) [ut(t); obj.branches{j}.v_ref(t)^2;...
                    obj.branches{j}.rhog(t);...
                    obj.branches{j}.rhoc(t);...
                    obj.branches{j}.qc(t)];
            end
        end
        
        function sigs = make_sigs(obj,t_ser,var)
            sigs.t=t_ser;
            sigs.all=var;
            sigs.vs=var(:,obj.idx_all.vs);
            sigs.P=var(:,obj.idx_all.P);
            sigs.Q=var(:,obj.idx_all.Q);
            sigs.q=var(:,obj.idx_all.q);
            sigs.xhat=var(:,obj.idx_all.xhat);
            sigs.zhat=var(:,obj.idx_all.zhat);
            
            ut_fun = obj.make_input;
            nt = numel(t_ser);
            sigs.ut = zeros(nt,size(ut_fun(0),1));
            w = zeros(nt,obj.nall);
            for i=1:numel(t_ser)
                ut_t=ut_fun(t_ser(i));
                sigs.ut(i,:)=transpose(ut_t);
                dists = obj.dis_man.discon_res.t;
                for j=1:numel(dists)
                    if t_ser(i)>=dists(j)
                        discon=obj.dis_man.discon_res.discon(:,j);
                    end
                end
                w(i,:)=transpose(obj.compute_what_linear(t_ser(i),transpose(var(i,:)),ut_t,discon));
            end
            idx_x=obj.obs_man.idx.x;
            idx_y=obj.obs_man.idx.y;
            idx_xh=obj.obs_man.idx.xhat;
            sigs.w = [var(:,idx_y) var(:,idx_x)];
            sigs.what = [w(:,idx_y) var(:,idx_xh)];
            sigs.vshat = w(:,obj.idx_all.vs);
            
            sigs.q_e = var(:,idx_x)-sigs.xhat;
            sigs.q_e_nom = sigs.q_e./sigs.q_e(end,:);
        end
        
        function u_refs = set_input(obj)
            ut = obj.make_input();
            u_refs = ut(0);
        end
        
        function [value,isterminal,direction] = DisEventFcn_detect(obj,t,y,u,a,discon)
            obsm=obj.obs_man;
            obss=obsm.obss;
            dism=obj.dis_man;
            d_obss=dism.discon_obss;
            nbs=numel(obj.branches);
            value=ones(nbs,1);
            isterminal=ones(nbs,1);
            direction=zeros(nbs,1);
            
            for i=1:numel(obss)
                obsi=obss{i};
                qtmp=y(obsi.idx.x);
                xhattmp=y(obsi.idx.xhat);
                res_tmp=abs(qtmp-xhattmp);
                if ~d_obss.connection(i) && any(res_tmp>d_obss.interval)
                    value(cell2mat(obsi.ob_branches_area))=0;
                end
            end
            
            % calculating estimated interconnection signals under attack and disconnection
            %{
            a_vec=a.attack_vec;
            a_t=a.t;
            u_at = add_attack(t,u,a_vec,a_t);
            y_dis = obj.dis_man.discon_var(y,discon);
            u_at_dis = obj.dis_man.discon_u(u_at,discon);
            w = obj.compute_what_linear(t,y_dis,u_at_dis,discon);
            
            for i=1:numel(obss)
                vtmp=y(obss{i}.idx.ym);
                wtmp=w(obss{i}.idx.ym);
                res(i)=abs(vtmp(1)-wtmp(1));
                if ~d_obss.connection(i) && res(i)>d_obss.interval
                    value(i)=0;
                end
            end
            %}
        end
        
        function [value,isterminal,direction] = DisEventFcn_time(obj,t,y)
            nb=numel(obj.branches);
            value=ones(nb,1);
            eps=1e-4;
            t0=obj.net_para.d_para.d_t-eps;
            if t0<t && t<t0+obj.t_int
                value(obj.net_para.d_para.d_sys)=0;
            end
            isterminal=ones(nb,1);
            direction=zeros(nb,1);
        end
        
        function xqh = sub_xqxh(obj,x,q,xhat,zhat)
            xqh=zeros(obj.nall,1);
            i_q=1; i_xhat=1; i_zhat=1; i_ot=1;
            for i=1:obj.nall
                if any(ismember(obj.idx_q,i))
                    xqh(i)=q(i_q);
                    i_q=i_q+1;
                elseif any(ismember(obj.obs_man.idx.xhat,i))
                    xqh(i)=xhat(i_xhat);
                    i_xhat=i_xhat+1;
                elseif any(ismember(obj.obs_man.idx.zhat,i))
                    xqh(i)=zhat(i_zhat);
                    i_zhat=i_zhat+1;
                else
                    xqh(i)=x(i_ot);
                    i_ot=i_ot+1;
                end
            end
        end


    end
end

function g2 = voltage_drop_eq(var)
    g2 = var(1)+var(2)+var(3);
    % var(1): v'_i+1^2
    % var(2): v'_i^2
    % var(3): 2beta_i+1(P'_i+1,Q'_i+1)
end

function b = beta_PQ(R,X,P,Q)
b=2*(R*P+X*Q);
end

function out = emp_to_zero(inp)
if isempty(inp)
    out=zeros(size(inp,1),1);
else
    out=inp;
end
end

function [a_v, a_t] = set_attack_zero(nu)
a_v=zeros(nu,1);
a_t=Inf;
end

function u = add_attack(t,u_refs,attack_vec,attack_t)
if t>attack_t
    u = u_refs+attack_vec;
else
    u = u_refs;
end
end