classdef obs_man < handle
    
    properties
        n_obstates_all
        obss
        net
        obs_C
        obs_B % \hat{w} = obs_C*x+obs_B*u,
        % x includes \hat{x}, \hat{w}, does 'not' include \hat{q]
        idx % idx.x,idx.y,idx.ot,idx.xhat,idx.xhat_ot
        PnP=1; % 1:PnP observer, 0:standard
    end
    
    methods
        function obj = obs_man(obs_tab,net)
            obj.net = net;
            nobs = height(obs_tab);
            obss = cell(nobs,1);
            nall = net.nyall+net.nxall;
            idx_obstates=nall;
            for i=1:nobs
                obss{i} = observer(obs_tab(i,:),net.branches);
                obsi=obss{i};
                
                indvec=cell2mat(obsi.ob_branches_area);
                idx_tmp.x=[];idx_tmp.y=[];idx_tmp.xhat=[];
                idx_tmp.ym=[];idx_tmp.ym_iny=[];

                for j=1:numel(indvec)
                    idx_vec = indvec(j);
                    bri = obj.net.branches{idx_vec};
                    idx_tmp.x = [idx_tmp.x;bri.idx.q];
                    idx_tmp.y = [idx_tmp.y;bri.idx.vs;bri.idx.P;bri.idx.Q];

                    if any(ismember(obsi.meas_brs,idx_vec))
                        idx_tmp.ym = [idx_tmp.ym;bri.idx.vs];
                    end
                end
                idx_tmp.ot=compute_idx_others(sort([idx_tmp.x;idx_tmp.y]),nall);
                idx_xhat_tmp = transpose(idx_obstates+1:idx_obstates+obsi.nx);
                idx_tmp.xhat = [idx_tmp.xhat; idx_xhat_tmp];
                
                if obj.PnP==1
                    obsi.nzhat=obsi.nx;
                    idx_tmp.zhat = transpose(idx_obstates+obsi.nx+1:idx_obstates+obsi.nx+obsi.nzhat);
                else
                    obsi.nzhat=0;
                    idx_tmp.zhat=[];
                end

                for j=1:numel(idx_tmp.ym)
                    idx_tmp.ym_iny=[idx_tmp.ym_iny;find([idx_tmp.y]==idx_tmp.ym(j))];
                end
                obss{i}.sub_idx(idx_tmp);
                
                model_mats = get_model_mats(obj,obsi);
                obsi.add_model_mats(model_mats);
                obsi.set_obs_gain(net.net_para.error_feedback);
                obsi.add_obsys_mats();
                obsi.w_dyn=make_w_dyn(obsi);
                
                idx_obstates=idx_obstates+numel(idx_tmp.xhat)+numel(idx_tmp.zhat);
            end
            obj.obss = obss;
            obj.n_obstates_all = idx_obstates-nall;
            
            [obj.obs_C,obj.obs_B] = make_CB(obj,idx_obstates);
            get_idx_obsm(obj);
        end
        
        function model_mats= get_model_mats(obj,obs)
            t = 0;
            nvar = obj.net.nxall+obj.net.nyall;
            x0 = zeros(nvar,1);
            %u0 = zeros(obs.nu,1);
            nv0 = 1;
            nv_ref = obj.net.num_DGfo;
            nu = nv0+nv_ref+3*obj.net.num_DGfo;
            u0 = zeros(nu,1);
            
            %{
            %% removed because of huge numerical error
            thx = 1e-7*ones(length(x0),1);
            thu = 1e-7*ones(length(u0),1);
            A = numjac(@(t,x) obj.get_Mf(obs,x,obj.net.v_ref(0)),0,x0,obj.get_Mf(obs,x0,obj.net.v_ref(0)),thx,[],0);
            B = numjac(@(t,u) obj.get_Mf(obs,x0,u),0,u0,obj.get_Mf(obs,x0,u0),thu,[],0);
            C = numjac(@(t,x) obj.get_Mg(obs,x,obj.net.v_ref(0)),0,x0,obj.get_Mg(obs,x0,obj.net.v_ref(0)),thx,[],0);
            D = numjac(@(t,u) obj.get_Mg(obs,x0,u),0,u0,obj.get_Mg(obs,x0,u0),thu,[],0);
            %}
            
            I = eye(nvar);
            %Iu = eye(obs.nu);
            Iu = eye(nu);
            Aall=[]; Ball=[]; Call=[]; Dall=[]; h=1e-6;
            for i=1:nvar
                e = I(:,i);
                Aall = [Aall, (obj.get_Mf(obs,t,x0+h*e,u0)-obj.get_Mf(obs,t,x0,u0))/h];
                Call = [Call, (obj.get_Mg(obs,t,x0+h*e,u0)-obj.get_Mg(obs,t,x0,u0))/h];
            end
            
            %for i=1:obs.nu
            for i=1:nu
                eu = Iu(:,i);
                Ball = [Ball, (obj.get_Mf(obs,t,x0,u0+h*eu)-obj.get_Mf(obs,t,x0,u0))/h];
                Dall = [Dall, (obj.get_Mg(obs,t,x0,u0+h*eu)-obj.get_Mg(obs,t,x0,u0))/h];
            end
            
            %
            % x: states in the area
            % y: outputs in the area
            % z: others
            % dx = Ax+Ly+Zz+Bu
            % 0  = Cx+Ey+Rz+Du
            % letting iE=-pinv(E),
            % y = iE*Cx+iE*Rz+iE*Du
            % dx = (A+L*iE*C)x+(Z+L*iE*R)z+(B+L*iE*D)u
            % -> dx = Ax+Lz+Bu
            % -> y  = Cx+Ez+Du
            %
            
            idx_x=obs.idx.x;idx_y=obs.idx.y;idx_ot=obs.idx.ot;
            
            A1=Aall(:,idx_x); L1=Aall(:,idx_y); Z1=Aall(:,idx_ot);
            C1=Call(:,idx_x); E1=Call(:,idx_y); R1=Call(:,idx_ot); 
            
            iE1 = -inv(E1);
            A=A1+L1*iE1*C1; L=Z1+L1*iE1*R1; B=Ball+L1*iE1*Dall;
            C=iE1*C1; E=iE1*R1; D=iE1*Dall;

            model_mats.A=A; model_mats.L=L; model_mats.B=B;
            model_mats.C=C; model_mats.E=E; model_mats.D=D;
            
            % y_meas = C_mx+E_mz+D_mu
            model_mats.Cm=C(obs.idx.ym_iny,:); model_mats.Em=E(obs.idx.ym_iny,:);
            model_mats.Dm=D(obs.idx.ym_iny,:);
        end
        
        function Mf = get_Mf(obj,obs,t,x,u)
            area = cell2mat(obs.ob_branches_area);
            I = eye(obj.net.nxall);
            M = [];
            nx = 0;
            for i=1:numel(obj.net.branches)
                if any(ismember(area,i)) && obj.net.branches{i}.nx>=1
                    M = [M; I(nx+1:nx+obj.net.branches{i}.nx,:)];
                end
                nx = nx+obj.net.branches{i}.nx;
            end
            Mf = M*obj.net.get_f(t,x,u);
        end
        
        function Mg = get_Mg(obj,obs,t,x,u)
            M = get_MforO(obs,obj.net.energy_flow,obj.net.vol_flow);
            Mg = M*obj.net.get_g(t,x,u);
        end
        
        function [C,B] = make_CB(obj,nall)
            % y=C[x;z]+Bu
            C=[]; B=[];
            nob = numel(obj.obss);
            idx_sort=[];
            for i=1:nob
                obsi = obj.obss{i};
                n = size(obsi.model_mats.C,1);
                C00 = zeros(n,nall);
                C00(:,obsi.idx.xhat)=obsi.model_mats.C;
                C00(:,obsi.idx.ot)=obsi.model_mats.E;
                if obj.PnP==1
                    C00(:,obsi.idx.zhat)=-obsi.model_mats.C;
                end
                idx_sort=[idx_sort;obsi.idx.y];
                
                C = [C;C00];
                B = [B;obsi.model_mats.D];
            end
            [~,I]=sort(idx_sort);
            C=C(I,:);B=B(I,:);
        end
        
        function get_idx_obsm(obj)
            idx_tmp.x=[]; idx_tmp.y=[];
            idx_tmp.xhat=[]; idx_tmp.zhat=[];
            for i=1:numel(obj.obss)
                obsi=obj.obss{i};
                idx_tmp.x=[idx_tmp.x;obsi.idx.x];
                idx_tmp.y=[idx_tmp.y;obsi.idx.y];
                idx_tmp.xhat=[idx_tmp.xhat;obsi.idx.xhat];
                idx_tmp.zhat=[idx_tmp.zhat;obsi.idx.zhat];
            end
            obj.idx.x=sort(idx_tmp.x); obj.idx.y=sort(idx_tmp.y);
            obj.idx.xhat=sort(idx_tmp.xhat); obj.idx.zhat=sort(idx_tmp.zhat);
            nxyall = obj.net.nxall+obj.net.nyall;
            obj.idx.ot=compute_idx_others(sort([idx_tmp.x;idx_tmp.y]),nxyall);
            nall = nxyall+obj.n_obstates_all;
            obj.idx.xzhat_ot=compute_idx_others(sort([idx_tmp.x;idx_tmp.y;idx_tmp.xhat;idx_tmp.zhat;obj.idx.ot]),nall);
        end
    end
end

function MforO = get_MforO(obs,energy_flow,vol_flow)
area = cell2mat(obs.ob_branches_area);
Mef = make_exMatrix(cell2mat(energy_flow),area);
Mvf = make_exMatrix(vol_flow,area);
MforO = blkdiag(Mef,Mef,Mvf);
end

function M = make_exMatrix(flow_mat,area)
M=[];
nflow=size(flow_mat,1);
I=eye(nflow);
for i=1:nflow
    if any(ismember(area,abs(flow_mat(i,2))))
        M=[M;I(i,:)];
    end
end
end

function idx_others = compute_idx_others(idx,n)
idx_others=[];
for i=1:n
    if ~any(ismember(idx,i))
        idx_others=[idx_others,i];
    end
end
idx_others=transpose(idx_others);
end

function Gam = make_w_dyn(obs)
idx_x=obs.idx.x;idx_y=obs.idx.y;idx_ot=obs.idx.ot;

C = obs.model_mats.C;
E = obs.model_mats.E;
D = obs.model_mats.D;
Gam = @(var,u) [C*var(idx_x)+E*var(idx_ot)+D*u;var(idx_x)];
end