classdef observer < handle
    
    properties
        nx
        ny
        nzhat
        ob_branches_area
        model_mats
        obsys_mats
        w_dyn
        idx % x,y,xhat,zhat,yall,ym,ym_iny
        obs_gain
        meas_brs % measured branches (voltage)
        % states: xhat,zhat
        % w: estimated interconnection signal
        % w_dyn: function \hat{w]_i = dyn(x,\hat{x},w,u)
    end
    
    methods
        
        function obj = observer(ob_para,branches)
            obj.ob_branches_area=ob_para{:,'branches_area'};
            obj.nx=0; obj.ny=0;
            indvec=cell2mat(obj.ob_branches_area);
            for i=1:numel(indvec)
                idx_vec = indvec(i);
                bri = branches{idx_vec};
                obj.nx = obj.nx+bri.nx;
                obj.ny = obj.ny+bri.ny;
            end
            obj.meas_brs=cell2mat(ob_para{:,'measured_branches'});
        end
        
        function add_model_mats(obj,mats)
            obj.model_mats=mats;
        end
        
        function set_obs_gain(obj,feedback)
            A_des=obj.model_mats.A;
            C_des=obj.model_mats.Cm;
            if feedback
                %q=4e1;
                q=1e7;
                Q=q*eye(obj.nx);R=eye(numel(obj.idx.ym_iny));
                lqr_sol=lqr(A_des',C_des',Q,R);
                obj.obs_gain=lqr_sol';
                if size(A_des,1)==1
                    obj.obs_gain=(A_des+3e-1)/C_des;
                end
            else
                obj.obs_gain=zeros(size(A_des,1),size(C_des,1));
            end
        end
        
        function add_obsys_mats(obj)
            obj.obsys_mats=obj.model_mats;
        end
        
        function sub_idx(obj,idx)
            obj.idx=idx;
        end
        
        function f = get_fi(obj,var,w,meas_sig,u_refs)
            % x1: states
            % w: interconnection from the others
            % u_refs: [v'_0^2;v_ref^2;rhog;rhoc;qc]
            
            f=zeros(obj.nx+obj.nzhat,1);
            xhat=var(1:obj.nx);zhat=var(obj.nx+1:obj.nx+obj.nzhat);
            A=obj.obsys_mats.A; L=obj.obsys_mats.L;
            B=obj.obsys_mats.B; C=obj.obsys_mats.Cm;
            E=obj.obsys_mats.Em; D=obj.obsys_mats.Dm;
            p=obj.obs_gain*(meas_sig-C*xhat-E*w-D*u_refs);
            
            f(1:obj.nx) = A*xhat+L*w+B*u_refs+p;
            if ~isempty(zhat)
                f(obj.nx+1:obj.nx+obj.nzhat) = A*zhat+p;
            end
            
            if size(A,1)==1
                %A-obj.obs_gain*C
                %meas_sig-D*u_refs
            end
            
        end
        
    end
    
end